module Parser where

import Data.Char(isLower, isDigit, digitToInt)
import Control.Applicative

-- Это просто отдельный файл для аппликативного парсера, который вы писали на лекции

newtype Parser tok a = Parser { runParser :: [tok] -> Maybe ([tok], a)}

instance Functor (Parser tok) where
    fmap g (Parser p) = Parser f where
        f xs = case p xs of
            Just (cs, c)    -> Just (cs, g c)
            Nothing         -> Nothing

instance Applicative (Parser tok) where
    -- pure :: a -> Parser tok a
    pure x = Parser $ (\s -> Just (s, x))
    -- (<*>) :: Parser tok (a -> b) -> Parser tok a -> Parser tok b
    Parser u <*> Parser v = Parser f where
        f xs = case u xs of
            Nothing         -> Nothing
            Just (xs', g)    -> case v xs' of
                Nothing         -> Nothing
                Just (xs'', x)  -> Just (xs'', g x)

instance Alternative (Parser tok) where
    -- empty :: Parser tok a
    empty = Parser $ \_ -> Nothing
    -- (<|>) :: Parser tok a -> Parser tok a -> Parser tok a
    Parser u <|> Parser v = Parser f where
        f xs = case u xs of
            Nothing -> v xs
            z       -> z


satisfy :: (tok -> Bool) -> Parser tok tok
satisfy pr = Parser f where
    f (t : ts) | pr t   = Just (ts, t)
    f _                 = Nothing

lower :: Parser Char Char
lower = satisfy isLower

char :: Char -> Parser Char Char
char c = satisfy (== c)

digit :: Parser Char Int
digit = digitToInt <$> satisfy isDigit

-- Напишите парсер, который парсит не просто одну цифру а число
nat :: Parser Char Int
nat = foldl1 (\num x -> num*10 + x) <$> some digit

-- Напишите парсер, результат работы которого -- заданная строка
-- runParser (string "ABC") "ABC123" = Just ("123", "ABC")
-- runParser (string "ABC") "CDE123" = Nothing
string :: String -> Parser Char String
string = traverse char
