module Sem09 where

import Data.Monoid((<>))
import Control.Applicative(ZipList(..), liftA2)

-- Немного про класс типов Traversable. Как вам уже рассказали на лекции,
-- он позволяет взять структуру данных и каким-либо образом ее обойти(название как бы намекает)
-- для этого у него есть два метода -- traverse и sequenceA(один, как вы видели,
-- выражается через другой и vice versa)
-- Тут могли бы быть примеры типа "чему равно значение выражения",
-- но как мы увидели на предыдущем занятии их очень стремно считать в уме
-- Поэтому, в качестве разминки, скажите пожалуйста:
--      0. sequenceA [(+1), (+2), (+3)] 3
--          sequenceA :: Applicative f => [f a] -> f [a]
--          sequenceA [] = pure []
--          sequenceA (x:xs) = (:) <$> x <*> sequenceA xs
--
--          sequenceA [(+1)] => (:) <$> (+1) <*> pure [] =>
--          (\x -> (:) . (+1) x []) 3 => [4]
--      1. Какой тип имеет traverse . traverse и что оно делает?
--      2. Что делает traverse mappend?

-- Напишите представителя Traversable для типа Result

data Result a = Ok a | Error String deriving (Eq, Show)

instance Functor Result where
    fmap g (Ok a) = Ok $ g a
    fmap _ (Error x) = Error x

instance Foldable Result where
    -- foldMap :: Monoid m => (a -> m) -> Result a -> m
    foldMap g (Ok x) = g x
    foldMap _ (Error _) = mempty

instance Traversable Result where
    -- traverse :: Applicative f => (a -> f b) -> Result a -> f (Result b)
    traverse g (Ok a) = Ok <$> g a
    traverse _ (Error x) = pure $ Error x

-- Заметьте, что выше нам потребовалось написать инстансы Foldable и Functor
-- Не очень-то и удивительно, если вспомнить определение класса Traversable
-- в котором и налагаются эти контексты. На лекции вам показывали, что можно
-- с помощью некоторого трюка реализовать методы Functor по умолчанию, воспользовавшись
-- тождественным функтором. Попробуем реализовать методы Foldable по умолчанию, воспользовавшись
-- функтором Const

newtype Const c a = Const { getConst :: c } deriving (Eq, Show)

-- Это так называемый фантомный тип. У конструктора типа _два_ типовых параметра,
-- а у конструктора данных - только один. Сделаем его функтором:

instance Functor (Const c) where
    fmap _ (Const v) = Const v

-- С точки зрения вычислений с эффектами -- это контейнер, в котором ничего не лежит(отображать нечего)
-- Но благодаря структуре функтора он умеет "перепаковывать" значения одного типа в другой. Поиграйтесь в GHCi
-- и убедитесь в этом. Тем временем представитель Foldable тоже не делает ничего интересного:

instance Foldable (Const c) where
    -- foldMap :: Monoid m => (a -> m) -> Const c a -> m
    foldMap _ _ = mempty

-- Осталось написать представителей Applicative и Traversable для нашего функтора

instance Monoid c => Applicative (Const c) where
    -- pure :: a -> Const c a
    pure _ = Const mempty
    -- (<*>) :: Const c (a -> b) -> Const c a -> Const c b
    Const f <*> Const x = Const (f <> x)

instance Traversable (Const c) where
    -- traverse :: (a -> f b) -> Const c a -> f (Const c b)
    traverse _ (Const v) = pure $ Const v

-- Теперь у нас есть все, чтобы написать реализацию foldMap по умолчанию

foldMapDefault :: (Traversable t, Monoid m) => (a -> m) -> t a -> m
foldMapDefault f = getConst . traverse (Const . f)

-- Понятно, что ничего не понятно, давайте посмотрим на типы и поймем, что убедимся, что все сходится
-- f :: a -> m
-- Const :: m -> Const m b
-- Const . f :: a -> Const m b
-- traverse (Const . f) :: t a -> Const m (t b)
-- getConst :: Const m a -> m
-- getConst . traverse (Const . f) :: t a -> m

-- См. parser.hs

-- Моноидальные функторы.

-- Рассмотрим следующий класс типов:

class Functor f => Monoidal f where
    unit :: f ()
    (*&*) :: f a -> f b -> f (a, b)

-- Чуть позже мы докажем, что он изоморфен аппликативному функтору, но пока давайте просто напишем
-- каких-нибудь его представителей

instance Monoidal [] where
    unit = [()]
    xs *&* ys = [(x, y) | x <- xs, y <- ys]

instance Monoidal ZipList where
    unit = ZipList $ repeat ()
    ZipList xs *&* ZipList ys = ZipList $ zip xs ys

-- Напишите представителя Monoidal для Maybe

instance Monoidal Maybe where
    unit = Just ()
    Just x *&* Just y = Just (x, y)
    _ *&* _ = Nothing

-- Для пары

instance Monoid m => Monoidal ((,) m) where
    unit = (mempty, ())
    (u, x) *&* (v, y) = (u <> v, (x, y))

-- Для (->) e

instance Monoidal ((->) e) where
    unit = const ()
    f *&* g = \x -> (f x, g x)

-- Теперь давайте выразим методы Monoidal через методы Applicative и vice versa

unit' :: Applicative f => f ()
unit' = pure ()

pair' :: Applicative f => f a -> f b -> f (a, b)
pair' u v = (,) <$> u <*> v

pure' :: Monoidal f => a -> f a
pure' x = const x <$> unit

ap' :: Monoidal f => f (a -> b) -> f a -> f b
ap' u v = uncurry id <$> u *&* v


-- Беззаконию не бывать, у моноидальных функторов есть свои законы, которым должен подчиняться
-- каждый представитель. Первые два:

-- 1. fst <$> (v *&* unit) = v (Right identity)
-- 2. snd <$> (unit *&* v) = v (Left identity)

-- Их мораль в том, что unit не делает вообще ничего интересного
-- Для остальных понадобятся вспомогательные комбинаторы:

(***) :: (a -> a') -> (b -> b') -> (a, b) -> (a', b')
f *** g = \(x, y) -> (f x, g y)

-- Очевидный изоморфизм между типами (a, (b, c)) и ((a, b), c)

asl :: (a, (b, c)) -> ((a, b), c)
asl (x, (y, z)) = ((x, y), z)

asr :: ((a, b), c) -> (a, (b, c))
asr ((x, y), z) = (x, (y, z))

-- Третий закон (Associativity)
-- asl <$> (u *&* (v *&* w)) = (u *&* v) *&* w

-- Четвертый (Naturality)
-- (f *** g) <$> (u *&* v) = (f <$> u) *&* (g <$> v)

-- Давайте, например, докажем, что аппликативные законы следуют из моноидальных
-- Для начала, докажем, что наше определение pure' и ap' удовлетворяет закону, который связывает
-- Applicative и Functor

-- f <$> v = pure' f `ap'` v (0)

-- Нам понадобится простенькая лемма:
-- uncurry id . (const f *** snd) = f . snd . snd (Lemma 1)
-- Докажите это утверждение:

-- Теперь докажем (0):
-- pure' f `ap'` v => (pure' def)
-- ap' (const f <$> unit) v => (ap' def)
-- uncurry id <$> ((const f <$> unit) *&* v) => (Left identity)
-- uncurry id <$> ((const f <$> unit) *&* (snd <$> (unit *&* v))) => (Naturality)
-- uncurry id <$> ((const f *** snd) <$> (unit *&* (unit *&* v))) => (Functor law)
-- (uncurry id . (const f *** snd)) <$> (unit *&* (unit *&* v)) => (Lemma 1)
-- (f . snd . snd) <$> (unit *&* (unit *&* v)) => (Functor law)
-- (f . snd) <$> (snd <$> (unit *&* (unit *&* v))) => (Left identity)
-- (f . snd) <$> (unit *&* v) => (Functor law)
-- f <$> (snd <$> (unit *&* v)) => (Left identity)
-- f <$> v

-- Очевидно, что pure id <*> v = v тривиально следует из предыдущего

-- Остальные законы доказываются аналогично, можете попробовать доказать самостоятельно
-- Однако композиция доказывается довольно сложно, будьте внимательны

-- Если вы дочитали до сюда и еще не упоролись, то вот статья Бартоша Милевски,
-- в которой он рассказывает, как связаны аппликатвные функторы из хаскеля и
-- теоретико-категориальные lax monoidal functors
-- https://bartoszmilewski.com/2017/02/06/applicative-functors/
